package org.processmining.xeslite.external;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.deckfour.xes.model.impl.XAttributeBooleanImpl;

import com.google.common.collect.Interner;

/**
 * This {@link MapDBAttributeMap} caches the three most commonly used attributes
 * concept:name, transition:lifecycle, org:resource and time:timestamp for fast
 * access in memory.
 * 
 * @author F. Mannhardt
 * 
 */
final class MapDBAttributeMapWithCache extends MapDBAttributeMap {

	private static final XAttribute UNCACHABLE = new XAttributeBooleanImpl("UNCACHEABLE", true);
	private static final XAttribute NULL = new XAttributeBooleanImpl("NULL", true);

	private final Interner<String> interner;

	public MapDBAttributeMapWithCache(ExternalAttributable owner, MapDBStore store, Interner<String> interner) {
		super(owner, store);
		this.interner = interner;
	}

	private final void cacheAttribute(String key, XAttribute value) {
		if (XConceptExtension.KEY_NAME.equals(key) && value instanceof XAttributeLiteral) {
			if (!value.hasAttributes()) {
				((AttributesCacheable) getOwner()).setName(interner.intern(((XAttributeLiteral) value).getValue()));
			}
		} else if (XLifecycleExtension.KEY_TRANSITION.equals(key) && value instanceof XAttributeLiteral) {
			if (!value.hasAttributes()) {
				((AttributesCacheable) getOwner())
						.setTransition(interner.intern(((XAttributeLiteral) value).getValue()));
			}
		} else if (XTimeExtension.KEY_TIMESTAMP.equals(key) && value instanceof XAttributeTimestamp) {
			if (!value.hasAttributes()) {
				((AttributesCacheable) getOwner()).setTime(((XAttributeTimestamp) value).getValueMillis());
			}
		}
	}

	private final XAttribute getFromCache(String key) {
		if (XConceptExtension.KEY_NAME.equals(key)) {
			return ((AttributesCacheable) getOwner()).getName() == null ? NULL : new XAttributeLiteralExternalImpl(
					getStore().getAttributeKeyPool().getIndex(XConceptExtension.KEY_NAME),
					((AttributesCacheable) getOwner()).getName(), XConceptExtension.instance(), getStore());
		} else if (XLifecycleExtension.KEY_TRANSITION.equals(key)) {
			return ((AttributesCacheable) getOwner()).getTransition() == null ? NULL
					: new XAttributeLiteralExternalImpl(getStore().getAttributeKeyPool().getIndex(
							XLifecycleExtension.KEY_TRANSITION), ((AttributesCacheable) getOwner()).getTransition(),
							XLifecycleExtension.instance(), getStore());
		} else if (XTimeExtension.KEY_TIMESTAMP.equals(key)) {
			return ((AttributesCacheable) getOwner()).getTime() == -1 ? NULL : new XAttributeTimestampExternalImpl(
					getStore().getAttributeKeyPool().getIndex(XTimeExtension.KEY_TIMESTAMP),
					((AttributesCacheable) getOwner()).getTime(), XTimeExtension.instance(), getStore());
		}
		return UNCACHABLE;
	}

	@Override
	public final XAttribute put(String key, XAttribute value) {
		if (getOwner() instanceof AttributesCacheable) {
			cacheAttribute(key, value);
		}
		// Need to also put into the underlying storage, as we do not provide cached answers for the iterators
		return super.put(key, value);
	}

	public final XAttribute get(Object key) {
		//TODO make sure that we are in sync for deletion
		if (getOwner() instanceof AttributesCacheable) {
			XAttribute cachedAttribute = getFromCache((String) key);
			if (cachedAttribute == NULL) {
				return null;
			} else if (cachedAttribute == UNCACHABLE) {
				return super.get(key);
			} else {
				return cachedAttribute;
			}
		} else {
			return super.get(key);
		}
	}

	public final XAttribute remove(Object key) {
		if (getOwner() instanceof AttributesCacheable) {
			removeFromCache(key);	
		}	
		return super.remove(key);
	}

	private final void removeFromCache(Object key) {
		if (XConceptExtension.KEY_NAME.equals(key)) {
			((AttributesCacheable) getOwner()).setName(null);
		} else if (XLifecycleExtension.KEY_TRANSITION.equals(key)) {
			((AttributesCacheable) getOwner()).setTransition(null);
		} else if (XTimeExtension.KEY_TIMESTAMP.equals(key)) {
			((AttributesCacheable) getOwner()).setTime(-1);
		}
	}

	public final void clear() {
		clearCache();
		super.clear();
	}

	private final void clearCache() {
		if (getOwner() instanceof AttributesCacheable) {
			((AttributesCacheable) getOwner()).setName(null);
			((AttributesCacheable) getOwner()).setTransition(null);
			((AttributesCacheable) getOwner()).setTime(-1);
		}
	}

}