/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */
package org.processmining.xeslite.external;

import java.net.URI;
import java.util.Date;

import org.deckfour.xes.extension.XExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.id.XID;
import org.deckfour.xes.model.XAttributable;
import org.deckfour.xes.model.XAttributeBoolean;
import org.deckfour.xes.model.XAttributeContainer;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeID;
import org.deckfour.xes.model.XAttributeList;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeMapImpl;
import org.mapdb.BTreeMap;
import org.mapdb.DB;

/**
 * Factory storing attributes of {@link XAttributable} XES objects in a,
 * possibly external, {@link ExternalStore}. This implementation stores
 * attributes of {@link XEvent}, {@link XTrace} objects in a {@link DiskStore}
 * on the file system using MapDB as storage provider. Attributes are stored
 * together in a single {@link BTreeMap}.
 * <p>
 * Use one of the embedded sub-classes in this class to create a version that
 * fits your needs. Currently, there are:
 * <ul>
 * <li> {@link MapDBDiskImpl}
 * <li> {@link MapDBDiskWithoutCacheImpl}
 * <li> {@link MapDBDiskSequentialAccessImpl}
 * <li> {@link MapDBDiskSequentialAccessWithoutCacheImpl}
 * </ul>
 * <p>
 * PLEASE NOTE: Attributes are immutable in this implementation, as they may be
 * serialized and re-created at any time. Trying to change the value of an
 * attribute for example through the {@link XAttributeLiteral#setValue(String)}
 * method will result in an {@link UnsupportedOperationException}. To change the
 * value of an existing attribute, please just use the
 * {@link XAttributeMap#put(String, org.deckfour.xes.model.XAttribute)} method.
 * 
 * @author F. Mannhardt
 * 
 */
public abstract class XFactoryExternalStore implements XFactory {

	/**
	 * Use 32 MB of memory assuming that each XAttribute takes up at most 144
	 * bytes, which would be approx. the case for {@link XAttributeLiteral} with
	 * a String of length 50.
	 */
	public static final int DEFAULT_CACHE_SIZE = (int) (16l * 1024l * 1024l) / 144;
	public static final long MIN_MEMORY_FOR_CACHING = 256l * 1024l * 1024l;
	

	/**
	 * A XES Factory that stores XAttributes with MapDB on disk.
	 * <p>
	 * This version does store common attributes like 'concept:name',
	 * 'lifecycle:transition', 'time:timestamp' in memory. This implementation
	 * is NOT guaranteed to be synchronized!
	 * 
	 * @author F. Mannhardt
	 *
	 */
	public static class MapDBDiskImpl extends XFactoryExternalStore {

		public static void register() {
			XFactoryRegistry.instance().register(new MapDBDiskImpl());
		}

		private final ExternalStore attributeStore;

		public MapDBDiskImpl() {
			this(new MapDBStore.Builder());
		}

		public MapDBDiskImpl(DB db) {
			this(new MapDBStore.Builder().withDB(db));
		}
		
		public MapDBDiskImpl(MapDBStore.Builder dbBuilder) {
			super();
			attributeStore = dbBuilder.build();
		}

		@Override
		protected final ExternalStore getStore() {
			return attributeStore;
		}

		@Override
		public String getName() {
			return "XESLite: MapDB Optimized for speed";
		}

		@Override
		public String getDescription() {
			return "A XES Factory that stores XAttributes with MapDB on disk. This version does store common attributes like 'concept:name', 'lifecycle:transition', 'time:timestamp' in memory. This implementation is NOT guaranteed to be synchronized!";
		}

	}

	/**
	 * A XES Factory that stores XAttributes with MapDB on disk. This version
	 * does <b>NOT</b> store common attributes like 'concept:name',
	 * 'lifecycle:transition', 'time:timestamp' in memory.
	 * 
	 * @author F. Mannhardt
	 *
	 */
	public static class MapDBDiskWithoutCacheImpl extends MapDBDiskImpl {

		public static void register() {
			XFactoryRegistry.instance().register(new MapDBDiskWithoutCacheImpl());
		}

		@Override
		public String getName() {
			return "XESLite: MapDB Optimized for memory";
		}

		@Override
		public String getDescription() {
			return "A XES Factory that stores XAttributes with MapDB on disk. This version does NOT store common attributes like 'concept:name', 'lifecycle:transition', 'time:timestamp' in memory. This implementation is NOT guaranteed to be synchronized!";
		}

		public XEvent createEvent() {
			return new XEventBareExternalImpl(getStore());
		}

		public XEvent createEvent(XAttributeMap attributes) {
			return new XEventBareExternalImpl(attributes, getStore());
		}

	}

	/**
	 * A XES Factory that stores XAttributes with MapDB on disk. This version is
	 * optimized for sequential access of events, performance will degraded when
	 * using random-access methods of XTrace.
	 * <p>
	 * This version does store common attributes like 'concept:name',
	 * 'lifecycle:transition', 'time:timestamp' in memory.
	 * 
	 * @author F. Mannhardt
	 *
	 */
	public static class MapDBDiskSequentialAccessImpl extends MapDBDiskImpl {

		public static void register() {
			XFactoryRegistry.instance().register(new MapDBDiskSequentialAccessImpl());
		}

		@Override
		public String getName() {
			return "XESLite: MapDB Optimized for speed & sequential access";
		}

		@Override
		public String getDescription() {
			return "A XES Factory that stores XAttributes with MapDB on disk. This version is optimized for sequential access of events, performance will degraded when using random-access methods of XTrace.  This version does store common attributes like 'concept:name', 'lifecycle:transition', 'time:timestamp' in memory. This implementation is NOT guaranteed to be synchronized!";
		}

		public XTrace createTrace() {
			return new XTraceCompressedExternalImpl(true, getStore());
		}

		public XTrace createTrace(XAttributeMap attributes) {
			return new XTraceCompressedExternalImpl(true, attributes, getStore());
		}

		public XEvent createEvent() {
			return new XEventCachingExternalImpl(getStore());
		}

		public XEvent createEvent(XAttributeMap attributes) {
			return new XEventCachingExternalImpl(attributes, getStore());
		}

	}

	/**
	 * A XES Factory that stores XAttributes with MapDB on disk. This version is
	 * optimized for sequential access of events, performance will degraded when
	 * using random-access methods of XTrace.
	 * <p>
	 * This version does <b>NOT</b> store common attributes like 'concept:name',
	 * 'lifecycle:transition', 'time:timestamp' in memory.
	 * 
	 * @author F. Mannhardt
	 *
	 */
	public static class MapDBDiskSequentialAccessWithoutCacheImpl extends MapDBDiskSequentialAccessImpl {

		public static void register() {
			XFactoryRegistry.instance().register(new MapDBDiskSequentialAccessWithoutCacheImpl());
		}

		@Override
		public String getName() {
			return "XESLite: MapDB Optimized for memory & sequential access";
		}

		@Override
		public String getDescription() {
			return "A XES Factory that stores XAttributes with MapDB on disk. This version is optimized for sequential access of events, performance will degraded when using random-access methods of XTrace. This version does NOT store common attributes like 'concept:name', 'lifecycle:transition', 'time:timestamp' in memory. This implementation is NOT guaranteed to be synchronized!";
		}
		
		public XTrace createTrace() {
			return new XTraceCompressedExternalImpl(false, getStore());
		}
		
		public XTrace createTrace(XAttributeMap attributes) {
			return new XTraceCompressedExternalImpl(false, attributes, getStore());
		}

		public XEvent createEvent() {
			return new XEventBareExternalImpl(getStore());
		}

		public XEvent createEvent(XAttributeMap attributes) {
			return new XEventBareExternalImpl(attributes, getStore());
		}

	}

	protected abstract ExternalStore getStore();

	private final Integer intern(String key) {
		return getStore().getAttributeKeyPool().put(key);
	}

	@Override
	public String toString() {
		return getName();
	}

	@Override
	public String getAuthor() {
		return "F. Mannhardt";
	}

	@Override
	public URI getUri() {
		return URI.create("http://www.xes-standard.org/");
	}

	@Override
	public String getVendor() {
		return "xes-standard.org";
	}

	@Override
	public XLog createLog() {
		return new XLogExternalImpl(getStore());
	}

	@Override
	public XLog createLog(XAttributeMap attributes) {
		return new XLogExternalImpl(attributes, getStore());
	}

	@Override
	public XTrace createTrace() {
		return new XTraceExternalImpl(new XAttributeMapImpl(), getStore());
	}

	@Override
	public XTrace createTrace(XAttributeMap attributes) {
		return new XTraceExternalImpl(attributes, getStore());
	}

	@Override
	public XEvent createEvent() {
		return new XEventCachingExternalImpl(getStore());
	}

	@Override
	public XEvent createEvent(XAttributeMap attributes) {
		return new XEventCachingExternalImpl(attributes, getStore());
	}

	@Override
	public XEvent createEvent(XID id, XAttributeMap attributes) {
		throw new UnsupportedOperationException("Can not create an XEvent with a pre-defined ID");
	}

	@Override
	public XAttributeMap createAttributeMap() {
		return new XAttributeMapImpl();
	}

	@Override
	public XAttributeBoolean createAttributeBoolean(String key, boolean value, XExtension extension) {
		return new XAttributeBooleanExternalImpl(intern(key), value, extension, getStore());
	}

	@Override
	public XAttributeContinuous createAttributeContinuous(String key, double value, XExtension extension) {
		return new XAttributeContinuousExternalImpl(intern(key), value, extension, getStore());
	}

	@Override
	public XAttributeDiscrete createAttributeDiscrete(String key, long value, XExtension extension) {
		return new XAttributeDiscreteExternalImpl(intern(key), value, extension, getStore());
	}

	@Override
	public XAttributeLiteral createAttributeLiteral(String key, String value, XExtension extension) {
		return new XAttributeLiteralExternalImpl(intern(key), value, extension, getStore());
	}

	@Override
	public XAttributeTimestamp createAttributeTimestamp(String key, Date value, XExtension extension) {
		return new XAttributeTimestampExternalImpl(intern(key), value, extension, getStore());
	}

	@Override
	public XAttributeTimestamp createAttributeTimestamp(String key, long millis, XExtension extension) {
		return new XAttributeTimestampExternalImpl(intern(key), millis, extension, getStore());
	}

	@Override
	public XAttributeID createAttributeID(String key, XID value, XExtension extension) {
		return new XAttributeIDExternalImpl(intern(key), value, extension, getStore());
	}

	@Override
	public XAttributeContainer createAttributeContainer(String key, XExtension extension) {
		return new XAttributeContainerExternalImpl(intern(key), extension, getStore());
	}

	@Override
	public XAttributeList createAttributeList(String key, XExtension extension) {
		return new XAttributeListExternalImpl(intern(key), extension, getStore());
	}

	public void commit() {
		getStore().commit();
	}

}
