package org.processmining.xeslite.query;

import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.regex.Pattern;

import org.deckfour.xes.model.XAttributable;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeBoolean;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.util.XAttributeUtils;
import org.processmining.xeslite.query.attributes.DateAttributeIndex;
import org.processmining.xeslite.query.attributes.DoubleAttributeIndex;
import org.processmining.xeslite.query.attributes.LiteralAttributeIndex;
import org.processmining.xeslite.query.syntax.ParseException;
import org.processmining.xeslite.query.syntax.QueryParser;
import org.processmining.xeslite.query.syntax.QueryParserVisitor;
import org.processmining.xeslite.query.syntax.QueryRoot;

import com.googlecode.cqengine.ConcurrentIndexedCollection;
import com.googlecode.cqengine.IndexedCollection;
import com.googlecode.cqengine.attribute.Attribute;
import com.googlecode.cqengine.index.Index;
import com.googlecode.cqengine.index.radixinverted.InvertedRadixTreeIndex;
import com.googlecode.cqengine.query.Query;
import com.googlecode.cqengine.query.QueryFactory;
import com.googlecode.cqengine.resultset.ResultSet;

class XIndexedEventsImpl implements XIndexedEvents {

	private final IndexedCollection<XEvent> indexedCollection;

	XIndexedEventsImpl(Collection<XEvent> events) {
		indexedCollection = new ConcurrentIndexedCollection<>();
		indexedCollection.addIndex(InvertedRadixTreeIndex.onAttribute(eventName()));
		// Bulk-add
		indexedCollection.addAll(events);
	}

	public void addIndex(Index<XEvent> index) {
		indexedCollection.addIndex(index);
	}

	public ResultSet<XEvent> retrieve(Query<XEvent> query) {
		return indexedCollection.retrieve(query, QueryFactory.noQueryOptions());
	}

	@SuppressWarnings("unchecked")
	public Query<XEvent> parseQuery(String queryString, AttributeTypeResolver attributeTypeResolver)
			throws ParseException {
		if (!queryString.isEmpty() && queryString.charAt(0) == '~') {
			if (queryString.length() > 1) {
				queryString = queryString.substring(1);
				return QueryFactory.matchesRegex(eventName(), queryString);	
			} else {
				throw new ParseException("Missing regular expression afterr '~'");
			}
		} else if (!queryString.isEmpty()) {
			QueryParser queryParser = new QueryParser(queryString);
			try {
				QueryRoot rootNode = queryParser.parse();
				QueryParserVisitor visitor = new QueryBuilderVisitor<XEvent>(this, attributeTypeResolver);
				return (Query<XEvent>) rootNode.jjtAccept(visitor, null);
			} catch (ParseException e) {
				Pattern queryPattern = Pattern.compile("(=|>|<|~|<=|>=|\\!=|and|or)", Pattern.CASE_INSENSITIVE);
				if (!queryPattern.matcher(queryString).find()) {
					// Simple query without any operator -> search in names
					return QueryFactory.equal(eventName(), queryString);
				} else {
					throw e;
				}
			}
		} else {
			// Fall back empty query
			return QueryFactory.all(XEvent.class);
		}
	}

	public Query<XEvent> parseQuery(String queryString) throws ParseException {
		return parseQuery(queryString, new AttributeTypeResolver() {

			public Class<? extends XAttribute> getAttributeType(final String attributeName) {
				// simply return the first class found
				for (XEvent e : indexedCollection) {
					XAttribute attribute = e.getAttributes().get(attributeName);
					if (attribute != null) {
						return XAttributeUtils.getType(attribute);
					}
				}
				return XAttributeLiteral.class;

			}
		});
	}

	public ResultSet<XEvent> retrieve(String queryString) throws ParseException {
		return indexedCollection.retrieve(parseQuery(queryString), QueryFactory.noQueryOptions());
	}

	public boolean remove(XEvent element) {
		return indexedCollection.remove(element);
	}

	public void add(XEvent element) {
		indexedCollection.add(element);
	}

	/**
	 * @return Attribute concept:name of the event
	 */
	public final Attribute<XEvent, String> eventName() {
		return stringAttribute("concept:name");	
	}

	/**
	 * String attribute for {@link XAttributeLiteral} and
	 * {@link XAttributeBoolean}
	 * 
	 * @param attributeName
	 * @return
	 */
	public Attribute<XEvent, String> stringAttribute(String attributeName) {
		return new LiteralAttributeIndex<XEvent>(XEvent.class, attributeName) {

			protected Iterable<XAttributable> getAttributables(XEvent event) {
				return Collections.<XAttributable>singleton(event);
			}

		};
	}

	/**
	 * Double attribute for {@link XAttributeContinuous} and
	 * {@link XAttributeDiscrete}
	 * 
	 * @param attributeName
	 * @return
	 */
	public Attribute<XEvent, Double> doubleAttribute(String attributeName) {
		return new DoubleAttributeIndex<XEvent>(XEvent.class, attributeName) {

			protected Iterable<XAttributable> getAttributables(XEvent event) {
				return Collections.<XAttributable>singleton(event);
			}

		};
	}

	/**
	 * Date attribute
	 * 
	 * @param attributeName
	 * @return
	 */
	public Attribute<XEvent, Date> dateAttribute(String attributeName) {
		return new DateAttributeIndex<XEvent>(XEvent.class, attributeName) {

			protected Iterable<XAttributable> getAttributables(XEvent event) {
				return Collections.<XAttributable>singleton(event);
			}

		};
	}

}
